#include <arc/jobftpcontrol.h>
#include <arc/notify.h>

#include <globus_ftp_control.h>

#include <libintl.h>
#define _(A) dgettext("arclib", ((A)))


JobFTPControl::JobFTPControl() throw(FTPControlError) : FTPControl() { }


JobFTPControl::~JobFTPControl() { }


std::string JobFTPControl::Submit(const URL& url,
                                  const std::string& rsl,
                                  int timeout,
                                  bool disconnectafteruse)
throw(JobFTPControlError, FTPControlError) {

	// dump rsl into temporary file
	std::string filename("/tmp/rsl.XXXXXX");
	int fd = mkstemp((char*)filename.c_str());
	if (fd==-1)
		throw JobFTPControlError(_("Could not create temporary file"));
	write(fd, (void*)rsl.c_str(), rsl.size());
	close(fd);

	std::string resp;

	try {
		Connect(url, timeout);
		resp = SendCommand("CWD " + url.Path(), timeout);
		resp = SendCommand("CWD new", timeout);
	} catch(FTPControlError e) {
		throw JobFTPControlError(e.what());
	}

	std::string::size_type pos;
	if ((pos = resp.find('"')) == std::string::npos)
		throw JobFTPControlError(_("Could not parse server response"));
	resp = resp.substr(pos+1);

	if ((pos = resp.find('"')) == std::string::npos)
		throw JobFTPControlError(_("Could not parse server response"));
	jobid = resp.substr(0, pos);

	pos = jobid.rfind('/');
	if (pos == std::string::npos)
		throw JobFTPControlError(_("Invalid jobid returned by server"));
	jobid = jobid.substr(pos+1);

	std::string uploadurl = url.str();
	if (uploadurl[uploadurl.length()-1] == '/') // remove trailing slash
		uploadurl.resize(uploadurl.length()-1);

	Upload(filename, uploadurl + "/new/job", timeout, false);
	if (disconnectafteruse) Disconnect(url, timeout);

	unlink(filename.c_str());

	return uploadurl + "/" + jobid;
}


std::string JobFTPControl::Submit(const URL& url,
                                  const std::string& rsl,
                                  std::map<std::string, std::string>& files,
                                  int timeout,
                                  bool disconnectafteruse)
throw(JobFTPControlError, FTPControlError) {

	Connect(url, timeout);

	std::string uploadurl = url.str();
	if (uploadurl[uploadurl.length()-1] == '/') // remove trailing slash
		uploadurl.resize(uploadurl.length()-1);

	// Submit RSL and obtain ID
	Submit(URL(uploadurl), rsl, timeout, false);

	// Submit local files
	std::map<std::string, std::string>::iterator fileit;
	for (fileit = files.begin(); fileit != files.end(); fileit++) {

		notify(DEBUG) << _("Uploading local file from") << " " << fileit->first
		              << " " << _("to") << " " << fileit->second <<std::endl;

		try {
			Upload(fileit->first,
			       uploadurl + "/" + jobid + "/" + fileit->second,
			       timeout,
			       false);
		} catch(FTPControlError e) {
			notify(WARNING) << e.what() << std::endl;
		}
	}

	if (disconnectafteruse) Disconnect(url, timeout);

	return uploadurl + "/" + jobid;
}


void JobFTPControl::Cancel(const std::string& jobid,
                           int timeout,
                           bool disconnectafteruse)
throw(JobFTPControlError, FTPControlError) {

	URL url(jobid);
	std::string path = url.Path();

	std::string::size_type pos = path.rfind('/');
	if (pos==std::string::npos || pos==0)
		throw JobFTPControlError(_("Illegal jobid specified"));

	std::string jobpath = path.substr(0, pos);
	std::string jobidnum = path.substr(pos+1);

	Connect(url, timeout);
	SendCommand("CWD " + jobpath, timeout);
	SendCommand("DELE " + jobidnum, timeout);
	if (disconnectafteruse) Disconnect(url, timeout);
}


void JobFTPControl::Clean(const std::string& jobid,
                          int timeout,
                          bool disconnectafteruse)
throw(JobFTPControlError, FTPControlError) {

	URL url(jobid);
	std::string path = url.Path();

	std::string::size_type pos = path.rfind('/');
	if (pos==std::string::npos || pos==0)
		throw JobFTPControlError(_("Illegal jobid specified"));

	std::string jobpath = path.substr(0, pos);
	std::string jobidnum = path.substr(pos+1);

	Connect(url, timeout);
	SendCommand("CWD " + jobpath, timeout);
	SendCommand("RMD " + jobidnum, timeout);
	if (disconnectafteruse) Disconnect(url, timeout);
}


void JobFTPControl::Resume(const std::string& jobid,
                           int timeout,
                           bool disconnectafteruse)
throw(JobFTPControlError, FTPControlError) {

	URL url(jobid);
	std::string urlstr = url.str();

	std::string::size_type pos = urlstr.rfind('/');
	if (pos==std::string::npos || pos==0)
		throw JobFTPControlError(_("Illegal jobid specified"));
	std::string jobnr = urlstr.substr(pos+1);
	urlstr = urlstr.substr(0, pos) + "/new/action";

	std::string rsl("&(action=restart)(jobid=" + jobnr + ")");

	// dump rsl into temporary file
	std::string filename("/tmp/rsl.XXXXXX");
	int fd = mkstemp((char*)filename.c_str());
	if (fd==-1)
		throw JobFTPControlError(_("Could not create temporary file"));
	write(fd, (void*)rsl.c_str(), rsl.size());
	close(fd);

	try {
		Connect(url, timeout);
		Upload(filename, urlstr, timeout, false);
		if (disconnectafteruse) Disconnect(url, timeout);
	} catch(FTPControlError e) {
		unlink(filename.c_str());
		throw;
	}

	unlink(filename.c_str());
}


void JobFTPControl::RenewCreds(const std::string& jobid,
                               int timeout,
                               bool disconnectafteruse)
throw(JobFTPControlError, FTPControlError) {

	URL url(jobid);
	std::string path = url.Path();

	std::string::size_type pos = path.rfind('/');
	if (pos==std::string::npos || pos==0)
		throw JobFTPControlError(_("Illegal jobid specified"));

	std::string jobpath = path.substr(0, pos);
	std::string jobidnum = path.substr(pos+1);

	Connect(url, timeout);
	SendCommand("CWD " + jobpath, timeout);
	SendCommand("CWD " + jobidnum, timeout);
	if (disconnectafteruse) Disconnect(url, timeout);
}


void CancelJob(const std::string& jobid)
throw(JobFTPControlError, FTPControlError) {
	JobFTPControl ftpc;
	ftpc.Cancel(jobid);
}


void CleanJob(const std::string& jobid)
throw(JobFTPControlError, FTPControlError) {
	JobFTPControl ftpc;
	ftpc.Clean(jobid);
}


void ResumeJob(const std::string& jobid)
throw(JobFTPControlError, FTPControlError) {
	JobFTPControl ftpc;
	ftpc.Resume(jobid);
}


void RenewCreds(const std::string& jobid)
throw(JobFTPControlError, FTPControlError) {
	JobFTPControl ftpc;
	ftpc.RenewCreds(jobid);
}

