#include <fstream>

#include <arc/configcore.h>
#include <arc/xmlconfig.h>

#include <libxml/parser.h>
#include <libxml/parserInternals.h>
#include <libxml/tree.h>
#include <libxml/xpath.h>

#include <libintl.h>
#define _(A) dgettext("arclib", ((A)))


static void XmlErrorHandler(void* ctx, const char* msg) {
	return;
}


void XMLConfig::fill_tree(xmlNode* node, Config& config) {

	bool has_element = false;

	for (xmlNode* cur_node = node; cur_node; cur_node = cur_node->next) {
		if (cur_node->type == XML_ELEMENT_NODE)
			has_element = true;
		if (cur_node->children)
			fill_tree(cur_node->children, config);
	}

	if (has_element)
		return;

	std::string key;
	for (xmlNode* cur_node = node; cur_node->parent->type != XML_DOCUMENT_NODE;
	     cur_node = cur_node->parent) {
		if (cur_node->type == XML_ELEMENT_NODE) {

			// Adding the attribute as an element
			for (xmlAttr *attribute = cur_node->properties; attribute;
			     attribute = attribute->next) {
				if (strcmp((char*) attribute->name, "label") == 0)
					if (key.empty())
						key = (char*) attribute->children->content;
					else
						key = (char*) attribute->children->content +
							('/' + key);
			}
			if (key.empty())
				key = (char*) cur_node->name;
			else
				key = (char*) cur_node->name + ('/' + key);
		}
	}
	config.AddConfValue(key, (char*) node->content);
}

Config XMLConfig::Read(std::istream& is) {

	Config config;

	// Create a parser context
	xmlParserCtxtPtr ctxt = xmlNewParserCtxt();
	if (ctxt == NULL)
		throw ConfigError(_("Failed to create parser context"));

	// Read stream content into buffer
	is.seekg(0, std::ios::end);
	std::streamsize length = is.tellg();
	is.seekg(0, std::ios::beg);
	char* buffer = new char[length + 1];
	is.read(buffer, length);
	buffer[length] = '\0';

	// Set Error-handler
	xmlSetGenericErrorFunc(NULL, (xmlGenericErrorFunc)XmlErrorHandler);

	// Parse and validate buffer and fill the tree
	xmlDocPtr doc;
	doc = xmlParseMemory(buffer, length);
	delete [] buffer;

	// Reset Error-handler
	xmlSetGenericErrorFunc(NULL, NULL);

	// Check if parsing suceeded
	if (doc == NULL) {
		xmlFreeParserCtxt(ctxt);
		throw ConfigError(_("Failed xml parsing"));
	}
	// Check if validation suceeded
	else if (ctxt->valid == 0) {
		xmlFreeParserCtxt(ctxt);
		xmlFreeDoc(doc);
		throw ConfigError(_("Failed to validate xml"));
	}

	// Free up the parser context
	xmlFreeParserCtxt(ctxt);

	// Walk the tree
	xmlNode *root_element = xmlDocGetRootElement(doc);

	fill_tree(root_element, config);

	xmlFreeDoc(doc);

	return config;
}

void XMLConfig::Write(Config config, std::ostream& os) {

	xmlDocPtr doc = NULL;
	xmlDtdPtr dtd = NULL;
	xmlNodePtr root_node = NULL;
	xmlNodePtr node = NULL;

	std::string root_name = "arc";

	LIBXML_TEST_VERSION;

	// Create a new document, a node and set it as a root node
	doc = xmlNewDoc(BAD_CAST "1.0");
	root_node = xmlNewNode(NULL, BAD_CAST root_name.c_str());
	xmlDocSetRootElement(doc, root_node);

	// Create DTD declaration.
	dtd = xmlCreateIntSubset(doc,
	                         BAD_CAST root_name.c_str(),
	                         NULL,
	                         BAD_CAST "nordugrid.dtd");

	// Create the content
	for (Options::iterator key = config.GetOptions().begin();
	     key != config.GetOptions().end(); key++) {

		std::string::size_type pos1 = key->first.find('/');
		std::string::size_type pos2 = key->first.rfind('/');

		std::string section = key->first.substr(0, pos1);
		std::string element = key->first.substr(pos2 + 1);
		std::string attribute;
		if (pos1 != pos2)
			attribute = key->first.substr(pos1 + 1, pos2 - pos1 - 1);

		// Loop over values
		std::list<std::string>::iterator value;
		for (std::list<std::string>::iterator value = key->second.begin();
		     value != key->second.end(); value++) {

			std::string xpath = '/' + root_name + '/' + section;
			if (!attribute.empty())
				xpath += "[@label='" + attribute + "']";
			else
				xpath += "[not(@label)]";

			xmlXPathContextPtr context = xmlXPathNewContext(doc);
			xmlXPathObjectPtr result =
				xmlXPathEvalExpression((xmlChar*) xpath.c_str(), context);
			xmlXPathFreeContext(context);

			// Full hierarchy (no attributes) support
			if (xmlXPathNodeSetIsEmpty(result->nodesetval)){
				std::string::size_type pos = 0;
				std::string::size_type pos2;
				node = root_node;
				while ((pos2 = section.find('/', pos)) != std::string::npos) {
					node = xmlNewChild(node, NULL, BAD_CAST
					                   section.substr(pos, pos2 - pos).c_str(),
					                   NULL);
					pos = pos2 + 1;
				}
				node = xmlNewChild(node, NULL,
				                   BAD_CAST section.substr(pos).c_str(),
				                   NULL);
			}
			else
				node = result->nodesetval->nodeTab[0];

			xmlNewChild(node,
			            NULL,
			            BAD_CAST element.c_str(),
			            BAD_CAST value->c_str());
			if (!attribute.empty())
				xmlSetProp(node,
				           BAD_CAST "label",
				           BAD_CAST attribute.c_str());
		}
	}

	xmlChar *xmlbuff;
	int buffersize;
	xmlDocDumpFormatMemory(doc, &xmlbuff, &buffersize, 1);

	// Send the buffer to the stream
	os << xmlbuff;

	// Free the buffer
	xmlFree(xmlbuff);

	//  free the document
	xmlFreeDoc(doc);

	// Free variables allocated by parser
	xmlCleanupParser();
}
